from binance.lib.utils import get_uuid, purge_map
from binance.error import ParameterArgumentError


def ping_connectivity(self, id=None):
    """Test connectivity to the WebSocket API.

    Args:
        id (str): request id

    Message sent:

    .. code-block:: json

        {
            "id": "922bcc6e-9de8-440d-9e84-7c80933a8d0d",
            "method": "ping"
        }


    Response:

    .. code-block:: json

        {
            "id": "922bcc6e-9de8-440d-9e84-7c80933a8d0d",
            "status": 200,
            "result": {},
            "rateLimits": [{
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 1
            }]
        }


    Note: You can use regular WebSocket ping frames to test connectivity as well,
    WebSocket API will respond with pong frames as soon as possible.

    """

    if not id:
        id = get_uuid()

    payload = {"id": id, "method": "ping"}

    self.send(payload)


def server_time(self, id=None):
    """Check server time

    Args:
        id (str): request id

    Message sent:

    .. code-block:: json

        {
            "id": "187d3cb2-942d-484c-8271-4e2141bbadb1",
            "method": "time"
        }


    Response:

    .. code-block:: json

        {
            "id": "187d3cb2-942d-484c-8271-4e2141bbadb1",
            "status": 200,
            "result": {
                "serverTime": 1656400526260
            },
            "rateLimits": [{
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 1
            }]
        }


    Test connectivity to the WebSocket API and get the current server time.

    """

    if not id:
        id = get_uuid()

    payload = {"id": id, "method": "time"}

    self.send(payload)


def exchange_info(self, **kwargs):
    """Exchange information

    Query current exchange trading rules, rate limits, and symbol information.

    Note: For symbol, Upper case is required.

    Keyword Args:
        symbols (list): symbols to get info
        symbol (str): symbol to get info
        permissons (list): permissions to get info

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "exchangeInfo",
            "params": {
                "symbols": [
                    "BNBBTC"
                ]
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": {
            "timezone": "UTC",
            "serverTime": 1655969291181,
            "rateLimits": [{
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200
            },
            {
                "rateLimitType": "ORDERS",
                "interval": "SECOND",
                "intervalNum": 10,
                "limit": 50
            },
            {
                "rateLimitType": "ORDERS",
                "interval": "DAY",
                "intervalNum": 1,
                "limit": 160000
            },
            {
                "rateLimitType": "RAW_REQUESTS",
                "interval": "MINUTE",
                "intervalNum": 5,
                "limit": 6100
            }],
            "exchangeFilters": [],
            "symbols": [{
                "symbol": "BNBBTC",
                "status": "TRADING",
                "baseAsset": "BNB",
                "baseAssetPrecision": 8,
                "quoteAsset": "BTC",
                "quotePrecision": 8,
                "quoteAssetPrecision": 8,
                "baseCommissionPrecision": 8,
                "quoteCommissionPrecision": 8,
                "orderTypes": [
                    "LIMIT",
                    "LIMIT_MAKER",
                    "MARKET",
                    "STOP_LOSS_LIMIT",
                    "TAKE_PROFIT_LIMIT"
                ],
                "icebergAllowed": true,
                "ocoAllowed": true,
                "quoteOrderQtyMarketAllowed": true,
                "allowTrailingStop": true,
                "cancelReplaceAllowed": true,
                "isSpotTradingAllowed": true,
                "isMarginTradingAllowed": true,
                "filters": [{
                    "filterType": "PRICE_FILTER",
                    "minPrice": "0.00000100",
                    "maxPrice": "100000.00000000",
                    "tickSize": "0.00000100"
                },
                {
                    "filterType": "LOT_SIZE",
                    "minQty": "0.00100000",
                    "maxQty": "100000.00000000",
                    "stepSize": "0.00100000"
                }],
                "permissions": [
                    "SPOT",
                    "MARGIN",
                    "TRD_GRP_004"
                ],
                "defaultSelfTradePreventionMode": "NONE",
                "allowedSelfTradePreventionModes": [
                    "NONE"
                ]
            }]},
            "rateLimits": [{
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 10
            }]
        }
    """
    parameters = {**kwargs}

    parameters = purge_map(parameters)

    if len(parameters) > 1:
        raise ParameterArgumentError(
            "Only one of symbol, symbols or permissions is required."
        )

    payload = {"id": parameters.pop("id", get_uuid()), "method": "exchangeInfo"}

    self.send(payload)


def order_book(self, symbol: str, **kwargs):
    """Order book

    Args:
        symbol (str): symbol to get order book

    Keyword Args:
        limit (int): limit of order book

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "depth",
            "params": {
                "symbol": "BNBBTC",
                "limit": 10
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": {
                "lastUpdateId": 2731179239,
                "bids": [
                    [
                        "0.01379900",
                        "3.43200000"
                    ],
                    [
                        "0.01379800",
                        "3.24300000"
                    ],
                    [
                        "0.01379700",
                        "10.45500000"
                    ],
                    [
                        "0.01379600",
                        "3.82100000"
                    ],
                    [
                        "0.01379500",
                        "10.26200000"
                    ]
                ],
                "asks": [
                    [
                        "0.01380000",
                        "5.91700000"
                    ],
                    [
                        "0.01380100",
                        "6.01400000"
                    ],
                    [
                        "0.01380200",
                        "0.26800000"
                    ],
                    [
                        "0.01380300",
                        "0.33800000"
                    ],
                    [
                        "0.01380400",
                        "0.26800000"
                    ]
                ]
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }
    """
    parameters = {"symbol": symbol.upper(), **kwargs}

    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "depth",
        "params": parameters,
    }

    self.send(payload)


def recent_trades(self, symbol: str, **kwargs):
    """Recent trades

    Args:
        symbol (str): symbol to get recent trades

    Keyword Args:
        limit (int): limit of recent trades

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "trades.recent",
            "params": {
                "symbol": "BNBBTC",
                "limit": 10
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": [
                {
                    "id": 194686783,
                    "price": "0.01361000",
                    "qty": "0.01400000",
                    "quoteQty": "0.00019054",
                    "time": 1660009530807,
                    "isBuyerMaker": true,
                    "isBestMatch": true
                }
            ],
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }
    """

    parameters = {"symbol": symbol.upper(), **kwargs}

    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "trades.recent",
        "params": parameters,
    }

    self.send(payload)


def historical_trades(self, symbol: str, apiKey: str, **kwargs):
    """Historical trades

    Args:
        symbol (str): symbol to get historical trades

    Keyword Args:
        symbol (str): symbol to get historical trades
        fromId (int): trade id to fetch from
        limit (int): limit of historical trades

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "trades.historical",
            "params": {
                "symbol": "BNBBTC",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "fromId": 0,
                "limit": 1
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": [
                {
                    "id": 0,
                    "price": "0.01361000",
                    "qty": "0.01400000",
                    "quoteQty": "0.00019054",
                    "time": 1660009530807,
                    "isBuyerMaker": true,
                    "isBestMatch": true
                }
            ],
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }
    """
    parameters = {"symbol": symbol.upper(), "apiKey": apiKey, **kwargs}

    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "trades.historical",
        "params": parameters,
    }

    self.send(payload)


def aggregate_trades(self, symbol: str, **kwargs):
    """Aggregate trades

    Args:
        symbol (str): symbol to get aggregate trades

    Keyword Args:
        symbol (str): symbol to get aggregate trades
        fromId (int): trade id to fetch from
        limit (int): limit of aggregate trades
        startTime (int): start time to fetch from
        endTime (int): end time to fetch from

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "trades.aggregate",
            "params": {
                "symbol": "BNBBTC",
                "fromId": 0,
                "limit": 1
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": [
                {
                    "a": 0,
                    "p": "0.01361000",
                    "q": "0.01400000",
                    "f": 0,
                    "l": 0,
                    "T": 1660009530807,
                    "m": true,
                    "M": true
                }
            ],
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """
    parameters = {"symbol": symbol.upper(), **kwargs}

    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "trades.aggregate",
        "params": parameters,
    }

    self.send(payload)


def klines(self, symbol: str, interval: str, **kwargs):
    """Klines/candlesticks

    Args:
        symbol (str): symbol to get klines
        interval (str): interval of klines

    Keyword Args:
        startTime (int): start time to fetch from
        endTime (int): end time to fetch from
        limit (int): limit of klines

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "klines",
            "params": {
                "symbol": "BNBBTC",
                "interval": "1m",
                "startTime": 1655969280000,
                "limit": 1
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": [
                [
                    1660009530807,
                    "0.01361000",
                    "0.01361000",
                    "0.01361000",
                    "0.01361000",
                    "0.01400000",
                    1660009530807,
                    "0.00019054",
                    0,
                    "0.00000000",
                    "0.00000000",
                    "0"
                ]
            ],
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """
    parameters = {"symbol": symbol.upper(), "interval": interval, **kwargs}

    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "klines",
        "params": parameters,
    }

    self.send(payload)


def ui_klines(self, symbol: str, interval: str, **kwargs):
    """Klines/candlesticks for UI

    Args:
        symbol (str): symbol to get klines
        interval (str): interval of klines

    Keyword Args:
        startTime (int): start time to fetch from
        endTime (int): end time to fetch from
        limit (int): limit of klines

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "uiKlines",
            "params": {
                "symbol": "BNBBTC",
                "interval": "1m",
                "startTime": 1655969280000,
                "limit": 1
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": [
                [
                    1660009530807,
                    "0.01361000",
                    "0.01361000",
                    "0.01361000",
                    "0.01361000",
                    "0.01400000",
                    1660009530807,
                    "0.00019054",
                    0,
                    "0.00000000",
                    "0.00000000",
                    "0"
                ]
            ],
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """
    parameters = {"symbol": symbol.upper(), "interval": interval, **kwargs}

    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "uiKlines",
        "params": parameters,
    }

    self.send(payload)


def avg_price(self, symbol: str, **kwargs):
    """Current average price for a symbol

    Args:
        symbol (str): symbol to get average price

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "avgPrice",
            "params": {
                "symbol": "BNBBTC"
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": {
                "mins": 5,
                "price": "0.01361000"
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """
    parameters = {"symbol": symbol.upper(), **kwargs}

    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "avgPrice",
        "params": parameters,
    }

    self.send(payload)


def ticker_24hr(self, **kwargs):
    """24 hour rolling window price change statistics

    Keyword Args:
        symbol (str): symbol to get ticker 24hr
        symbols (list): symbols to get ticker 24hr
        type (str): type of ticker 24hr

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "ticker.24hr",
            "params": {
                "symbol": "BNBBTC"
            }
        }

    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": {
                "symbol": "BNBBTC",
                "priceChange": "-94.99999800",
                "priceChangePercent": "-95.960",
                "weightedAvgPrice": "0.29628482",
                "prevClosePrice": "0.10002000",
                "lastPrice": "4.00000200",
                "lastQty": "200.00000000",
                "bidPrice": "4.00000000",
                "bidQty": "10.00000000",
                "askPrice": "4.00000200",
                "askQty": "10.00000000",
                "openPrice": "99.00000000",
                "highPrice": "100.00000000",
                "lowPrice": "0.10000000",
                "volume": "8913.30000000",
                "quoteVolume": "15.30000000",
                "openTime": 1499783499040,
                "closeTime": 1499869899040,
                "firstId": 28385,
                "lastId": 28460,
                "count": 76
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """
    parameters = purge_map(kwargs)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "ticker.24hr",
        "params": parameters,
    }

    self.send(payload)


def ticker(self, **kwargs):
    """Rolling window price change statistics

    Keyword Args:
        symbol (str): symbol to get ticker
        symbols (list): symbols to get ticker
        type (str): type of ticker
        windowSize (str): window size of ticker

    Message sent:

    .. code-block:: json

        {
            "id": "f4b3b507-c8f2-442a-81a6-b2f12daa030f",
            "method": "ticker",
            "params": {
                "symbols": [
                    "BNBBTC",
                    "BTCUSDT"
                ],
                "windowSize": "7d"
            }
        }


    Response

    .. code-block:: json

        {
            "id": "f4b3b507-c8f2-442a-81a6-b2f12daa030f",
            "status": 200,
            "result": {
                "symbol": "BNBBTC",
                "priceChange": "0.00061500",
                "priceChangePercent": "4.735",
                "weightedAvgPrice": "0.01368242",
                "openPrice": "0.01298900",
                "highPrice": "0.01418800",
                "lowPrice": "0.01296000",
                "lastPrice": "0.01360400",
                "volume": "587179.23900000",
                "quoteVolume": "8034.03382165",
                "openTime": 1659580020000,
                "closeTime": 1660184865291,
                "firstId": 192977765,
                "lastId": 195365758,
                "count": 2387994
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 2
                }
            ]
        }

    """
    parameters = purge_map(kwargs)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "ticker",
        "params": parameters,
    }

    self.send(payload)


def ticker_price(self, **kwargs):
    """Symbol price ticker

    Keyword Args:
        symbol (str): symbol to get ticker price
        symbols (list): symbols to get ticker price

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "ticker.price",
            "params": {
                "symbol": "BNBBTC"
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": {
                "symbol": "BNBBTC",
                "price": "4.00000200"
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """
    parameters = purge_map(kwargs)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "ticker.price",
        "params": parameters,
    }

    self.send(payload)


def ticker_book(self, **kwargs):
    """Symbol order book ticker

    Keyword Args:
        symbol (str): symbol to get ticker book
        symbols (list): symbols to get ticker book

    Message sent:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "method": "ticker.book",
            "params": {
                "symbol": "BNBBTC"
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5494febb-d167-46a2-996d-70533eb4d976",
            "status": 200,
            "result": {
                "symbol": "BNBBTC",
                "bidPrice": "4.00000000",
                "bidQty": "431.00000000",
                "askPrice": "4.00000200",
                "askQty": "9.00000000"
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """
    parameters = purge_map(kwargs)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "ticker.book",
        "params": parameters,
    }

    self.send(payload)



def portfolio_margin_account(self, **kwargs):
    """Get Portfolio Margin Account Info (USER_DATA)

    GET /sapi/v1/portfolio/account

    https://binance-docs.github.io/apidocs/spot/en/#get-portfolio-margin-account-info-user_data

    Keyword Args:
      recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/portfolio/account", {**kwargs})


def portfolio_margin_collateral_rate(self):
    """Portfolio Margin Collateral Rate (MARKET_DATA)

    Portfolio Margin Collateral Rate.

    Weight(IP): 50

    GET /sapi/v1/portfolio/collateralRate

    https://binance-docs.github.io/apidocs/spot/en/#portfolio-margin-collateral-rate-market_data

    """

    url_path = "/sapi/v1/portfolio/collateralRate"
    return self.sign_request("GET", url_path)


def portfolio_margin_bankruptcy_loan_amount(self, **kwargs):
    """Query Portfolio Margin Bankruptcy Loan Amount (USER_DATA)

    Query Portfolio Margin Bankruptcy Loan Amount.

    Weight(UID): 500

    GET /sapi/v1/portfolio/pmLoan

    https://binance-docs.github.io/apidocs/spot/en/#query-portfolio-margin-bankruptcy-loan-amount-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    url_path = "/sapi/v1/portfolio/pmLoan"
    return self.sign_request("GET", url_path, {**kwargs})


def portfolio_margin_bankruptcy_loan_repay(self, **kwargs):
    """Portfolio Margin Bankruptcy Loan Repay (USER_DATA)

    Repay Portfolio Margin Bankruptcy Loan.

    Weight(UID): 3000

    POST /sapi/v1/portfolio/repay

    https://binance-docs.github.io/apidocs/spot/en/#portfolio-margin-bankruptcy-loan-repay-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    url_path = "/sapi/v1/portfolio/repay"
    return self.sign_request("POST", url_path, {**kwargs})



from binance.lib.utils import get_uuid, purge_map, websocket_api_signature


def new_order(self, symbol: str, side: str, type: str, **kwargs):
    """Create a new order.

    Args:
        symbol (str): Symbol to trade.
        side (str): Side of the trade.
        type (str): Type of the trade.
    Keyword Args:
        timeInForce (str): Time in force.
        quantity (float): Quantity of the trade.
        quoteOrderQty (float): Quote order quantity.
        price (float): Price of the trade.
        newClientOrderId (str): New client order id.
        stopPrice (float): Stop price.
        trailingDelta (int): Trailing delta.
        icebergQty (float): Iceberg quantity.
        strategyId (int): Strategy id.
        strategyType (int): Strategy type.
        selfTradePreventionMode (str): Self trade prevention.
        newOrderRespType (str): New order response type.
        recvWindow (int): Recv window.

    Message sent:

    .. code-block:: json

        {
            "id": "56374a46-3061-486b-a311-99ee972eb648",
            "method": "order.place",
            "params": {
                "symbol": "BTCUSDT",
                "side": "SELL",
                "type": "LIMIT",
                "timeInForce": "GTC",
                "price": "23416.10000000",
                "quantity": "0.00847000",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "15af09e41c36f3cc61378c2fbe2c33719a03dd5eba8d0f9206fbda44de717c88",
                "timestamp": 1660801715431
            }
        }



    Response

    .. code-block:: json

        {
            "id": "56374a46-3061-486b-a311-99ee972eb648",
            "status": 200,
            "result": {
                "symbol": "BTCUSDT",
                "orderId": 12569099453,
                "orderListId": -1,
                "clientOrderId": "4d96324ff9d44481926157ec08158a40",
                "transactTime": 1660801715639
            },
            "rateLimits": [
                {
                    "rateLimitType": "ORDERS",
                    "interval": "SECOND",
                    "intervalNum": 10,
                    "limit": 50,
                    "count": 1
                },
                {
                    "rateLimitType": "ORDERS",
                    "interval": "DAY",
                    "intervalNum": 1,
                    "limit": 160000,
                    "count": 1
                },
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """

    parameters = {"symbol": symbol, "side": side, "type": type, **kwargs}
    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "order.place",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)


def new_order_test(self, symbol: str, side: str, type: str, **kwargs):
    """Test new order (TRADE)

    Args:
        symbol (str): Symbol to trade.
        side (str): Side of the trade.
        type (str): Type of the trade.
    Keyword Args:
        timeInForce (str): Time in force.
        quantity (float): Quantity of the trade.
        quoteOrderQty (float): Quote order quantity.
        price (float): Price of the trade.
        newClientOrderId (str): New client order id.
        stopPrice (float): Stop price.
        trailingDelta (int): Trailing delta.
        icebergQty (float): Iceberg quantity.
        strategyId (int): Strategy id.
        strategyType (int): Strategy type.
        selfTradePreventionMode (str): Self trade prevention.
        newOrderRespType (str): New order response type.
        recvWindow (int): Recv window.

    Message sent:

    .. code-block:: json

        {
            "id": "56374a46-3061-486b-a311-99ee972eb648",
            "method": "order.test",
            "params": {
                "symbol": "BTCUSDT",
                "side": "SELL",
                "type": "LIMIT",
                "timeInForce": "GTC",
                "price": "23416.10000000",
                "quantity": "0.00847000",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "15af09e41c36f3cc61378c2fbe2c33719a03dd5eba8d0f9206fbda44de717c88",
                "timestamp": 1660801715431
            }
        }



    Response

    .. code-block:: json

        {
            "id": "6ffebe91-01d9-43ac-be99-57cf062e0e30",
            "status": 200,
            "result": {},
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """

    parameters = {"symbol": symbol, "side": side, "type": type, **kwargs}
    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "order.test",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)


def get_order(self, symbol: str, **kwargs):
    """Get order (USER_DATA)

    Args:
        symbol (str): Symbol to trade.
    Keyword Args:
        orderId (int): Order id.
        origClientOrderId (str): Original client order id.
        recvWindow (int): Recv window.

    Message sent:

    .. code-block:: json

        {
            "id": "aa62318a-5a97-4f3b-bdc7-640bbe33b291",
            "method": "order.status",
            "params": {
                "symbol": "BTCUSDT",
                "orderId": 12569099453,
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "2c3aab5a078ee4ea465ecd95523b77289f61476c2f238ec10c55ea6cb11a6f35",
                "timestamp": 1660801720951
            }
        }

    Response:

    .. code-block:: json

        {
            "id": "aa62318a-5a97-4f3b-bdc7-640bbe33b291",
            "status": 200,
            "result": {
                "symbol": "BTCUSDT",
                "orderId": 12569099453,
                "orderListId": -1,
                "clientOrderId": "4d96324ff9d44481926157",
                "price": "23416.10000000",
                "origQty": "0.00847000",
                "executedQty": "0.00847000",
                "cummulativeQuoteQty": "198.33521500",
                "status": "FILLED",
                "timeInForce": "GTC",
                "type": "LIMIT",
                "side": "SELL",
                "stopPrice": "0.00000000",
                "trailingDelta": 10,
                "trailingTime": -1,
                "icebergQty": "0.00000000",
                "time": 1660801715639,
                "updateTime": 1660801717945,
                "isWorking": true,
                "workingTime": 1660801715639,
                "origQuoteOrderQty": "0.00000000",
                "strategyId": 37463720,
                "strategyType": 1000000,
                "selfTradePreventionMode": "NONE",
                "preventedMatchId": 0,
                "preventedQuantity": "1.200000"
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 2
                }
            ]
        }

    """

    parameters = {"symbol": symbol, **kwargs}
    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "order.status",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)


def cancel_order(self, symbol: str, **kwargs):
    """Cancel order (USER_DATA)
      Args:
        symbol (str): Symbol to trade.
      Keyword Args:
        orderId (int): Order id.
        origClientOrderId (str): Original client order id.
        newClientOrderId (str): New client order id.
        recvWindow (int): Recv window.

      Message sent:

    .. code-block:: json

        {
            "id": "5633b6a2-90a9-4192-83e7-925c90b6a2fd",
            "method": "order.cancel",
            "params": {
                "symbol": "BTCUSDT",
                "origClientOrderId": "4d96324ff9d44481926157",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "33d5b721f278ae17a52f004a82a6f68a70c68e7dd6776ed0be77a455ab855282",
                "timestamp": 1660801715830
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "5633b6a2-90a9-4192-83e7-925c90b6a2fd",
            "status": 200,
            "result": {
                "symbol": "BTCUSDT",
                "origClientOrderId": "4d96324ff9d44481926157",
                "orderId": 12569099453,
                "orderListId": -1,
                "clientOrderId": "91fe37ce9e69c90d6358c0",
                "price": "23416.10000000",
                "origQty": "0.00847000",
                "executedQty": "0.00001000",
                "cummulativeQuoteQty": "0.23416100",
                "status": "CANCELED",
                "timeInForce": "GTC",
                "type": "LIMIT",
                "side": "SELL",
                "stopPrice": "0.00000000",
                "trailingDelta": 0,
                "trailingTime": -1,
                "icebergQty": "0.00000000",
                "strategyId": 37463720,
                "strategyType": 1000000,
                "selfTradePreventionMode": "NONE"
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """

    parameters = {"symbol": symbol, **kwargs}
    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "order.cancel",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)


def cancel_replace_order(
    self, symbol: str, cancelReplaceMode: str, side: str, type: str, **kwargs
):
    """Cancel and replace order
    Args:
        symbol (str): Symbol to trade.
        cancelReplaceMode (str):
        side (str): SIDE
        type (str): ORDER_TYPE
    Keyword Args:
        cancelOrderId (int): Cancel order by orderId
        cancelOrigClientOrderId (int): Cancel order by clientOrderId.
        cancelNewClientOrderId (str): New ID for the canceled order. Automatically generated if not sent
        timeInForce (str): Time in force.
        price (float): Price.
        quantity (float): Quantity.
        quoteOrderQty (float): Quote order quantity.
        newClientOrderId (str): New client order id.
        newOrderRespType (str): New order response type.
        stopPrice (float): Stop price.
        trailingDelta (float): Trailing delta.
        icebergQty (float): Iceberg quantity.
        strategyId (int): Strategy id.
        strategyType (int): Strategy type.
        selfTradePreventionMode (str): Self trade prevention mode.
        recvWindow (int): Recv window.

    Message Sent:

    .. code-block:: json

        {
            "id": "99de1036-b5e2-4e0f-9b5c-13d751c93a1a",
            "method": "order.cancelReplace",
            "params": {
                "symbol": "BTCUSDT",
                "cancelReplaceMode": "ALLOW_FAILURE",
                "cancelOrigClientOrderId": "4d96324ff9d44481926157",
                "side": "SELL",
                "type": "LIMIT",
                "timeInForce": "GTC",
                "price": "23416.10000000",
                "quantity": "0.00847000",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "7028fdc187868754d25e42c37ccfa5ba2bab1d180ad55d4c3a7e2de643943dc5",
                "timestamp": 1660813156900
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "b220edfe-f3c4-4a3a-9d13-b35473783a25",
            "status": 409,
            "error": {
                "code": -2021,
                "msg": "Order cancel-replace partially failed.",
                "data": {
                "cancelResult": "SUCCESS",
                "newOrderResult": "FAILURE",
                "cancelResponse": {
                    "symbol": "BTCUSDT",
                    "origClientOrderId": "4d96324ff9d44481926157",
                    "orderId": 125690984230,
                    "orderListId": -1,
                    "clientOrderId": "91fe37ce9e69c90d6358c0",
                    "price": "23450.00000000",
                    "origQty": "0.00847000",
                    "executedQty": "0.00001000",
                    "cummulativeQuoteQty": "0.23450000",
                    "status": "CANCELED",
                    "timeInForce": "GTC",
                    "type": "LIMIT",
                    "side": "SELL",
                    "selfTradePreventionMode": "NONE"
                },
                "newOrderResponse": {
                    "code": -2010,
                    "msg": "Order would immediately match and take."
                }
                }
            },
            "rateLimits": [
                {
                    "rateLimitType": "ORDERS",
                    "interval": "SECOND",
                    "intervalNum": 10,
                    "limit": 50,
                    "count": 1
                },
                {
                    "rateLimitType": "ORDERS",
                    "interval": "DAY",
                    "intervalNum": 1,
                    "limit": 160000,
                    "count": 1
                },
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

        {
            "id": "ce641763-ff74-41ac-b9f7-db7cbe5e93b1",
            "status": 409,
            "error": {
                "code": -2021,
                "msg": "Order cancel-replace partially failed.",
                "data": {
                "cancelResult": "FAILURE",
                "newOrderResult": "SUCCESS",
                "cancelResponse": {
                    "code": -2011,
                    "msg": "Unknown order sent."
                },
                "newOrderResponse": {
                    "symbol": "BTCUSDT",
                    "orderId": 12569099453,
                    "orderListId": -1,
                    "clientOrderId": "bX5wROblo6YeDwa9iTLeyY",
                    "transactTime": 1660813156959,
                    "price": "23416.10000000",
                    "origQty": "0.00847000",
                    "executedQty": "0.00000000",
                    "cummulativeQuoteQty": "0.00000000",
                    "status": "NEW",
                    "timeInForce": "GTC",
                    "type": "LIMIT",
                    "side": "SELL",
                    "workingTime": 1669693344508,
                    "fills": [],
                    "selfTradePreventionMode": "NONE"
                }
                }
            },
            "rateLimits": [
                {
                    "rateLimitType": "ORDERS",
                    "interval": "SECOND",
                    "intervalNum": 10,
                    "limit": 50,
                    "count": 1
                },
                {
                    "rateLimitType": "ORDERS",
                    "interval": "DAY",
                    "intervalNum": 1,
                    "limit": 160000,
                    "count": 1
                },
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """

    parameters = {
        "symbol": symbol,
        "cancelReplaceMode": cancelReplaceMode,
        "side": side,
        "type": type,
        **kwargs,
    }
    parameters = purge_map(parameters)
    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "order.cancelReplace",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }
    self.send(payload)


def get_open_orders(self, **kwargs):
    """Current open orders (USER_DATA)

    Keyword Arguments:
        symbol (str): Symbol.
        recvWindow (int): Recv window.

    Message Sent:

    .. code-block:: json

        {
            "id": "55f07876-4f6f-4c47-87dc-43e5fff3f2e7",
            "method": "openOrders.status",
            "params": {
                "symbol": "BTCUSDT",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "d632b3fdb8a81dd44f82c7c901833309dd714fe508772a89b0a35b0ee0c48b89",
                "timestamp": 1660813156812
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "55f07876-4f6f-4c47-87dc-43e5fff3f2e7",
            "status": 200,
            "result": [
                {
                    "symbol": "BTCUSDT",
                    "orderId": 12569099453,
                    "orderListId": -1,
                    "clientOrderId": "4d96324ff9d44481926157",
                    "price": "23416.10000000",
                    "origQty": "0.00847000",
                    "executedQty": "0.00720000",
                    "cummulativeQuoteQty": "172.43931000",
                    "status": "PARTIALLY_FILLED",
                    "timeInForce": "GTC",
                    "type": "LIMIT",
                    "side": "SELL",
                    "stopPrice": "0.00000000",
                    "icebergQty": "0.00000000",
                    "time": 1660801715639,
                    "updateTime": 1660801717945,
                    "isWorking": true,
                    "workingTime": 1660801715639,
                    "origQuoteOrderQty": "0.00000000",
                    "selfTradePreventionMode": "NONE"
                }
            ],
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 3
                }
            ]
        }

    """

    parameters = purge_map(kwargs)
    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "openOrders.status",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }
    self.send(payload)


def cancel_open_orders(self, symbol: str, **kwargs):
    """Cancel all open orders on a symbol (USER_DATA)

    Arguments:
        symbol (str): Symbol.
    Keyword Arguments:
        recvWindow (int): Recv window.

    Message Sent:

    .. code-block:: json

        {
            "id": "778f938f-9041-4b88-9914-efbf64eeacc8",
            "method": "openOrders.cancelAll"
            "params": {
                "symbol": "BTCUSDT",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "773f01b6e3c2c9e0c1d217bc043ce383c1ddd6f0e25f8d6070f2b66a6ceaf3a5",
                "timestamp": 1660805557200
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "778f938f-9041-4b88-9914-efbf64eeacc8",
            "status": 200,
            "result": [
                {
                    "symbol": "BTCUSDT",
                    "origClientOrderId": "4d96324ff9d44481926157",
                    "orderId": 12569099453,
                    "orderListId": -1,
                    "clientOrderId": "91fe37ce9e69c90d6358c0",
                    "price": "23416.10000000",
                    "origQty": "0.00847000",
                    "executedQty": "0.00001000",
                    "cummulativeQuoteQty": "0.23416100",
                    "status": "CANCELED",
                    "timeInForce": "GTC",
                    "type": "LIMIT",
                    "side": "SELL",
                    "stopPrice": "0.00000000",
                    "trailingDelta": 0,
                    "trailingTime": -1,
                    "icebergQty": "0.00000000",
                    "strategyId": 37463720,
                    "strategyType": 1000000,
                    "selfTradePreventionMode": "NONE"
                },
                {
                    "orderListId": 19431,
                    "contingencyType": "OCO",
                    "listStatusType": "ALL_DONE",
                    "listOrderStatus": "ALL_DONE",
                    "listClientOrderId": "iuVNVJYYrByz6C4yGOPPK0",
                    "transactionTime": 1660803702431,
                    "symbol": "BTCUSDT",
                    "orders": [
                        {
                        "symbol": "BTCUSDT",
                        "orderId": 12569099453,
                        "clientOrderId": "bX5wROblo6YeDwa9iTLeyY"
                        },
                        {
                        "symbol": "BTCUSDT",
                        "orderId": 12569099454,
                        "clientOrderId": "Tnu2IP0J5Y4mxw3IATBfmW"
                        }
                    ],
                    "orderReports": [
                        {
                            "symbol": "BTCUSDT",
                            "origClientOrderId": "bX5wROblo6YeDwa9iTLeyY",
                            "orderId": 12569099453,
                            "orderListId": 19431,
                            "clientOrderId": "OFFXQtxVFZ6Nbcg4PgE2DA",
                            "price": "23450.50000000",
                            "origQty": "0.00850000",
                            "executedQty": "0.00000000",
                            "cummulativeQuoteQty": "0.00000000",
                            "status": "CANCELED",
                            "timeInForce": "GTC",
                            "type": "STOP_LOSS_LIMIT",
                            "side": "BUY",
                            "stopPrice": "23430.00000000",
                            "selfTradePreventionMode": "NONE"
                        },
                        {
                            "symbol": "BTCUSDT",
                            "origClientOrderId": "Tnu2IP0J5Y4mxw3IATBfmW",
                            "orderId": 12569099454,
                            "orderListId": 19431,
                            "clientOrderId": "OFFXQtxVFZ6Nbcg4PgE2DA",
                            "price": "23400.00000000",
                            "origQty": "0.00850000",
                            "executedQty": "0.00000000",
                            "cummulativeQuoteQty": "0.00000000",
                            "status": "CANCELED",
                            "timeInForce": "GTC",
                            "type": "LIMIT_MAKER",
                            "side": "BUY",
                            "selfTradePreventionMode": "NONE"
                        }
                    ]
                }
            ],
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """

    parameters = purge_map({"symbol": symbol, **kwargs})
    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "openOrders.cancelAll",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }
    self.send(payload)


def new_oco_order(self, symbol: str, side: str, price, quantity, **kwargs):
    """Place a new OCO Order (TRADE)
    Args:
        symbol (str): Symbol.
        side (str): BUY or SELL.
        price (float): Price.
        quantity (float): Quantity.
    Keyword Arguments:
        listClientOrderId (str): A unique id for the entire order list.
        limitClientOrderId (str): A unique id for the limit order.
        limitIcebergQty (float): Iceberg quantity.
        limitStrategyId (int): Strategy id.
        limitStrategyType (int): Strategy type.
        stopPrice (float): Stop price.
        trailingDelta (float): Trailing delta.
        stopClientOrderId (str): A unique id for the stop loss order.
        stopLimitPrice (float): Stop limit price.
        stopLimitTimeInForce (str): GTC or FOK.
        stopIcebergQty (float): Iceberg quantity.
        stopStrategyId (int): Strategy id.
        stopStrategyType (int): Strategy type.
        newOrderRespType (str): ACK, RESULT, or FULL; MARKET and LIMIT order types default to FULL, all other orders default to ACK.
        selfTradePreventionMode (str): NONE, EXPIRE_TAKER, EXPIRE_MAKER or EXPIRE_BOTH.
        recvWindow (int): Recv window.

    Message sent:

    .. code-block:: json

        {
            "id": "56374a46-3061-486b-a311-99ee972eb648",
            "method": "orderList.place",
            "params": {
                "symbol": "BTCUSDT",
                "side": "SELL",
                "price": "23420.00000000",
                "quantity": "0.00650000",
                "stopPrice": "23410.00000000",
                "stopLimitPrice": "23405.00000000",
                "stopLimitTimeInForce": "GTC",
                "newOrderRespType": "RESULT",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "6689c2a36a639ff3915c2904871709990ab65f3c7a9ff13857558fd350315c35",
                "timestamp": 1660801713767
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "57833dc0-e3f2-43fb-ba20-46480973b0aa",
            "status": 200,
            "result": {
                "orderListId": 1274512,
                "contingencyType": "OCO",
                "listStatusType": "EXEC_STARTED",
                "listOrderStatus": "EXECUTING",
                "listClientOrderId": "08985fedd9ea2cf6b28996",
                "transactionTime": 1660801713793,
                "symbol": "BTCUSDT",
                "orders": [
                    {
                        "symbol": "BTCUSDT",
                        "orderId": 12569138901,
                        "clientOrderId": "BqtFCj5odMoWtSqGk2X9tU"
                    },
                    {
                        "symbol": "BTCUSDT",
                        "orderId": 12569138902,
                        "clientOrderId": "jLnZpj5enfMXTuhKB1d0us"
                    }
                ],
                "orderReports": [
                    {
                        "symbol": "BTCUSDT",
                        "orderId": 12569138901,
                        "orderListId": 1274512,
                        "clientOrderId": "BqtFCj5odMoWtSqGk2X9tU",
                        "transactTime": 1660801713793,
                        "price": "23410.00000000",
                        "origQty": "0.00650000",
                        "executedQty": "0.00000000",
                        "cummulativeQuoteQty": "0.00000000",
                        "status": "NEW",
                        "timeInForce": "GTC",
                        "type": "STOP_LOSS_LIMIT",
                        "side": "SELL",
                        "stopPrice": "23405.00000000",
                        "workingTime": -1,
                        "selfTradePreventionMode": "NONE"
                    },
                    {
                        "symbol": "BTCUSDT",
                        "orderId": 12569138902,
                        "orderListId": 1274512,
                        "clientOrderId": "jLnZpj5enfMXTuhKB1d0us",
                        "transactTime": 1660801713793,
                        "price": "23420.00000000",
                        "origQty": "0.00650000",
                        "executedQty": "0.00000000",
                        "cummulativeQuoteQty": "0.00000000",
                        "status": "NEW",
                        "timeInForce": "GTC",
                        "type": "LIMIT_MAKER",
                        "side": "SELL",
                        "workingTime": 1660801713793,
                        "selfTradePreventionMode": "NONE"
                    }
                ]
            },
            "rateLimits": [
                {
                    "rateLimitType": "ORDERS",
                    "interval": "SECOND",
                    "intervalNum": 10,
                    "limit": 50,
                    "count": 2
                },
                {
                    "rateLimitType": "ORDERS",
                    "interval": "DAY",
                    "intervalNum": 1,
                    "limit": 160000,
                    "count": 2
                },
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """

    parameters = {
        "symbol": symbol,
        "side": side,
        "price": price,
        "quantity": quantity,
        **kwargs,
    }

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "orderList.place",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }
    self.send(payload)


def get_oco_order(self, **kwargs):
    """Get OCO order

    Keyword Args:
        orderListId (int): Order list id.
        origClientOrderId (str): Original client order id.
        recvWindow (int): Recv window.

    Message sent:

    .. code-block:: json

        {
            "id": "b53fd5ff-82c7-4a04-bd64-5f9dc42c2100",
            "method": "orderList.status",
            "params": {
                "symbol": "BTCUSDT",
                "origClientOrderId": "08985fedd9ea2cf6b28996"
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "d12f4e8892d46c0ddfbd43d556ff6d818581b3be22a02810c2c20cb719aed6a4",
                "timestamp": 1660801713965
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "b53fd5ff-82c7-4a04-bd64-5f9dc42c2100",
            "status": 200,
            "result": {
            "orderListId": 1274512,
            "contingencyType": "OCO",
            "listStatusType": "EXEC_STARTED",
            "listOrderStatus": "EXECUTING",
            "listClientOrderId": "08985fedd9ea2cf6b28996",
            "transactionTime": 1660801713793,
            "symbol": "BTCUSDT",
            "orders": [
                {
                    "symbol": "BTCUSDT",
                    "orderId": 12569138901,
                    "clientOrderId": "BqtFCj5odMoWtSqGk2X9tU"
                },
                {
                    "symbol": "BTCUSDT",
                    "orderId": 12569138902,
                    "clientOrderId": "jLnZpj5enfMXTuhKB1d0us"
                }
            ]
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 2
                }
            ]
        }

    """

    parameters = {**kwargs}

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "orderList.status",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }
    self.send(payload)


def cancel_oco_order(self, symbol: str, **kwargs):
    """Cancel OCO order

    Args:
        symbol (str): Symbol to place the order on.
    Keyword Args:
        orderListId (int): Order list id.
        listClientOrderId (str): List client order id.
        newClientOrderId (str): New client order id.
        recvWindow (int): Recv window.

    Message sent:

    .. code-block:: json

        {
            "id": "c5899911-d3f4-47ae-8835-97da553d27d0",
            "method": "orderList.cancel",
            "params": {
                "symbol": "BTCUSDT",
                "orderListId": 1274512,
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "4973f4b2fee30bf6d45e4a973e941cc60fdd53c8dd5a25edeac96f5733c0ccee",
                "timestamp": 1660801720210
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "c5899911-d3f4-47ae-8835-97da553d27d0",
            "status": 200,
            "result": {
                "orderListId": 1274512,
                "contingencyType": "OCO",
                "listStatusType": "ALL_DONE",
                "listOrderStatus": "ALL_DONE",
                "listClientOrderId": "6023531d7edaad348f5aff",
                "transactionTime": 1660801720215,
                "symbol": "BTCUSDT",
                "orders": [
                    {
                        "symbol": "BTCUSDT",
                        "orderId": 12569138901,
                        "clientOrderId": "BqtFCj5odMoWtSqGk2X9tU"
                    },
                    {
                        "symbol": "BTCUSDT",
                        "orderId": 12569138902,
                        "clientOrderId": "jLnZpj5enfMXTuhKB1d0us"
                    }
                ],
                "orderReports": [
                    {
                        "symbol": "BTCUSDT",
                        "orderId": 12569138901,
                        "orderListId": 1274512,
                        "clientOrderId": "BqtFCj5odMoWtSqGk2X9tU",
                        "transactTime": 1660801720215,
                        "price": "23410.00000000",
                        "origQty": "0.00650000",
                        "executedQty": "0.00000000",
                        "cummulativeQuoteQty": "0.00000000",
                        "status": "CANCELED",
                        "timeInForce": "GTC",
                        "type": "STOP_LOSS_LIMIT",
                        "side": "SELL",
                        "stopPrice": "23405.00000000",
                        "selfTradePreventionMode": "NONE"
                    },
                    {
                        "symbol": "BTCUSDT",
                        "orderId": 12569138902,
                        "orderListId": 1274512,
                        "clientOrderId": "jLnZpj5enfMXTuhKB1d0us",
                        "transactTime": 1660801720215,
                        "price": "23420.00000000",
                        "origQty": "0.00650000",
                        "executedQty": "0.00000000",
                        "cummulativeQuoteQty": "0.00000000",
                        "status": "CANCELED",
                        "timeInForce": "GTC",
                        "type": "LIMIT_MAKER",
                        "side": "SELL",
                        "selfTradePreventionMode": "NONE"
                    }
                ]
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 1
                }
            ]
        }

    """

    parameters = {"symbol": symbol, **kwargs}

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "orderList.cancel",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }
    self.send(payload)


def get_open_oco_orders(self, **kwargs):
    """Get open OCO orders

    Keyword Args:
        recvWindow (int): Recv window.

    Message sent:

    .. code-block:: json

        {
            "id": "3a4437e2-41a3-4c19-897c-9cadc5dce8b6",
            "method": "openOrderLists.status",
            "params": {
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "1bea8b157dd78c3da30359bddcd999e4049749fe50b828e620e12f64e8b433c9",
                "timestamp": 1660801713831
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "3a4437e2-41a3-4c19-897c-9cadc5dce8b6",
            "status": 200,
            "result": [
                {
                    "orderListId": 0,
                    "contingencyType": "OCO",
                    "listStatusType": "EXEC_STARTED",
                    "listOrderStatus": "EXECUTING",
                    "listClientOrderId": "08985fedd9ea2cf6b28996",
                    "transactionTime": 1660801713793,
                    "symbol": "BTCUSDT",
                    "orders": [
                        {
                            "symbol": "BTCUSDT",
                            "orderId": 4,
                            "clientOrderId": "CUhLgTXnX5n2c0gWiLpV4d"
                        },
                        {
                            "symbol": "BTCUSDT",
                            "orderId": 5,
                            "clientOrderId": "1ZqG7bBuYwaF4SU8CwnwHm"
                        }
                    ]
                }
            ],
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 3
                }
            ]
        }

    """

    parameters = {**kwargs}

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "openOrderLists.status",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }
    self.send(payload)



from binance.websocket.websocket_client import BinanceWebsocketClient


class SpotWebsocketAPIClient(BinanceWebsocketClient):
    def __init__(
        self,
        stream_url="wss://ws-api.binance.com/ws-api/v3",
        api_key=None,
        api_secret=None,
        on_message=None,
        on_open=None,
        on_close=None,
        on_error=None,
        on_ping=None,
        on_pong=None,
    ):
        self.api_key = api_key
        self.api_secret = api_secret

        super().__init__(
            stream_url,
            on_message=on_message,
            on_open=on_open,
            on_close=on_close,
            on_error=on_error,
            on_ping=on_ping,
            on_pong=on_pong,
        )

    # Market
    from binance.websocket.spot.websocket_api._market import ping_connectivity
    from binance.websocket.spot.websocket_api._market import server_time
    from binance.websocket.spot.websocket_api._market import exchange_info
    from binance.websocket.spot.websocket_api._market import order_book
    from binance.websocket.spot.websocket_api._market import recent_trades
    from binance.websocket.spot.websocket_api._market import historical_trades
    from binance.websocket.spot.websocket_api._market import aggregate_trades
    from binance.websocket.spot.websocket_api._market import klines
    from binance.websocket.spot.websocket_api._market import ui_klines
    from binance.websocket.spot.websocket_api._market import avg_price
    from binance.websocket.spot.websocket_api._market import ticker_24hr
    from binance.websocket.spot.websocket_api._market import ticker
    from binance.websocket.spot.websocket_api._market import ticker_price
    from binance.websocket.spot.websocket_api._market import ticker_book

    # Account
    from binance.websocket.spot.websocket_api._account import account
    from binance.websocket.spot.websocket_api._account import order_rate_limit
    from binance.websocket.spot.websocket_api._account import order_history
    from binance.websocket.spot.websocket_api._account import oco_history
    from binance.websocket.spot.websocket_api._account import my_trades
    from binance.websocket.spot.websocket_api._account import prevented_matches

    # Trade
    from binance.websocket.spot.websocket_api._trade import new_order
    from binance.websocket.spot.websocket_api._trade import new_order_test
    from binance.websocket.spot.websocket_api._trade import get_order
    from binance.websocket.spot.websocket_api._trade import cancel_order
    from binance.websocket.spot.websocket_api._trade import cancel_replace_order
    from binance.websocket.spot.websocket_api._trade import get_open_orders
    from binance.websocket.spot.websocket_api._trade import cancel_open_orders
    from binance.websocket.spot.websocket_api._trade import new_oco_order
    from binance.websocket.spot.websocket_api._trade import get_oco_order
    from binance.websocket.spot.websocket_api._trade import cancel_oco_order
    from binance.websocket.spot.websocket_api._trade import get_open_oco_orders

    # User Data Stream
    from binance.websocket.spot.websocket_api._user_data import user_data_start
    from binance.websocket.spot.websocket_api._user_data import user_data_ping
    from binance.websocket.spot.websocket_api._user_data import user_data_stop



from binance.lib.utils import check_required_parameters, check_required_parameter


def gift_card_create_code(self, token: str, amount: float, **kwargs):
    """Create a Binance Code (USER_DATA)

    POST /sapi/v1/giftcard/createCode

    This API is for creating a Binance Code. To get started with, please make sure:

    - You have a sufﬁcient balance in your Binance funding wallet
    - You need Enable Withdrawals for the API Key which requests this endpoint.

    https://binance-docs.github.io/apidocs/spot/en/#create-a-binance-code-user_data

    Args:
      token (str): The coin type contained in the Binance Code
      amount (float): The amount of the coin
    Keyword Args:
      recvWindow (int, optional): The value cannot be greater than 60000
    """
    params = [[token, "token"], [amount, "amount"]]
    check_required_parameters(params)

    payload = {"token": token, "amount": amount, **kwargs}

    return self.sign_request("POST", "/sapi/v1/giftcard/createCode", payload)


def gift_card_redeem_code(self, code: str, **kwargs):
    """Redeem a Binance Code (USER_DATA)

    POST /sapi/v1/giftcard/redeemCode

    This API is for redeeming the Binance Code. Once redeemed, the coins will be deposited in your funding wallet.

    Please note that if you enter the wrong code 5 times within 24 hours, you will no longer be able to redeem any Binance Code that day.

    https://binance-docs.github.io/apidocs/spot/en/#redeem-a-binance-code-user_data

    Args:
      code (str): Binance Code
    Keyword Args:
        externalUid (str, optional): Each external unique ID represents a unique user on the partner platform. The function helps you to identify the redemption behavior of different users, such as redemption frequency and amount. It also helps risk and limit control of a single account, such as daily limit on redemption volume, frequency, and incorrect number of entries. This will also prevent a single user account reach the partner's daily redemption limits. We strongly recommend you to use this feature and transfer us the User ID of your users if you have different users redeeming Binance codes on your platform. To protect user data privacy, you may choose to transfer the user id in any desired format (max. 400 characters).
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameter(code, "code")

    payload = {"code": code, **kwargs}

    return self.sign_request("POST", "/sapi/v1/giftcard/redeemCode", payload)


def gift_card_verify_code(self, referenceNo: str, **kwargs):
    """Verify a Binance Code (USER_DATA)

    GET /sapi/v1/giftcard/verify

    This API is for verifying whether the Binance Code is valid or not by entering Binance Code or reference number.

    Please note that if you enter the wrong binance code 5 times within an hour, you will no longer be able to verify any binance code for that hour.

    https://binance-docs.github.io/apidocs/spot/en/#verify-a-binance-code-user_data

    Args:
      referenceNo (str): reference number
    Keyword Args:
      recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameter(referenceNo, "referenceNo")

    payload = {"referenceNo": referenceNo, **kwargs}

    return self.sign_request("GET", "/sapi/v1/giftcard/verify", payload)


def gift_card_rsa_public_key(self, **kwargs):
    """Fetch RSA Public Key (USER_DATA)

    This API is for fetching the RSA Public Key.
    This RSA Public key will be used to encrypt the card code.
    Please note that the RSA Public key fetched is valid only for the current day.

    Weight(IP): 1

    GET /sapi/v1/giftcard/cryptography/rsa-public-key

    https://binance-docs.github.io/apidocs/spot/en/#fetch-rsa-public-key-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    url_path = "/sapi/v1/giftcard/cryptography/rsa-public-key"
    return self.sign_request("GET", url_path, {**kwargs})


def gift_card_buy_code(
    self, baseToken: str, faceToken: str, baseTokenAmount: float, **kwargs
):
    """Create a dual-token gift card (fixed value, discount feature) (TRADE)

    POST /sapi/v1/giftcard/buyCode

    https://binance-docs.github.io/apidocs/spot/en/#create-a-dual-token-gift-card-fixed-value-discount-feature-trade

    Args:
      baseToken (str): The coin type used to buy the Binance Code
      faceToken (str): The coin type contained in the Binance Code
      baseTokenAmount (float): The amount of the coin used to buy the Binance Code
    Keyword Args:
      discount (float, optional): The discount rate of the Binance Code
      recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameters(
        [
            [baseToken, "baseToken"],
            [faceToken, "faceToken"],
            [baseTokenAmount, "baseTokenAmount"],
        ]
    )

    payload = {
        "baseToken": baseToken,
        "faceToken": faceToken,
        "baseTokenAmount": baseTokenAmount,
        **kwargs,
    }

    return self.sign_request("POST", "/sapi/v1/giftcard/buyCode", payload)


def gift_card_token_limit(self, baseToken: str, **kwargs):
    """Get the limit of a token (TRADE)

    GET /sapi/v1/giftcard/buyCode/token-limit

    https://binance-docs.github.io/apidocs/spot/en/#fetch-token-limit-user_data

    Args:
      baseToken (str): The coin type used to buy the Binance Code
    Keyword Args:
      recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameter(baseToken, "baseToken")
    return self.sign_request(
        "GET",
        "/sapi/v1/giftcard/buyCode/token-limit",
        {"baseToken": baseToken, **kwargs},
    )



from binance.lib.utils import check_required_parameters


def bswap_pools(self):
    """List All Swap Pools (MARKET_DATA)
    Get metadata about all swap pools.

    GET /sapi/v1/bswap/pools

    https://binance-docs.github.io/apidocs/spot/en/#list-all-swap-pools-market_data

    """

    return self.limit_request("GET", "/sapi/v1/bswap/pools", {})


def bswap_liquidity(self, **kwargs):
    """Get liquidity information of a pool (USER_DATA)
    Get liquidity information and user share of a pool

    GET /sapi/v1/bswap/liquidity

    https://binance-docs.github.io/apidocs/spot/en/#get-liquidity-information-of-a-pool-user_data

    Keyword Args:
        poolId (int, optional)
        recvWindow (int, optional): The value cannot be greater than 60000

    """
    return self.sign_request("GET", "/sapi/v1/bswap/liquidity", kwargs)


def bswap_liquidity_add(self, poolId: int, asset: str, quantity: float, **kwargs):
    """Add Liquidity (TRADE)
    Add liquidity to a pool.

    POST /sapi/v1/bswap/liquidityAdd

    https://binance-docs.github.io/apidocs/spot/en/#add-liquidity-trade

    Args:
        poolId (int)
        asset (str)
        quantity (float)
    Keyword Args:
        type (str, optional): "Single" to add a single token; "Combination" to add dual tokens. Default "Single"
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameters(
        [[poolId, "poolId"], [asset, "asset"], [quantity, "quantity"]]
    )
    payload = {"poolId": poolId, "asset": asset, "quantity": quantity, **kwargs}

    return self.sign_request("POST", "/sapi/v1/bswap/liquidityAdd", payload)


def bswap_liquidity_remove(
    self, poolId: str, type: str, asset: list, shareAmount, **kwargs
):
    """Remove Liquidity (TRADE)
    Remove liquidity from a pool, type include SINGLE and COMBINATION, asset is mandatory for single asset removal

    POST /sapi/v1/bswap/liquidityRemove

    https://binance-docs.github.io/apidocs/spot/en/#remove-liquidity-trade

    Args:
        poolId (int)
        type (str): SINGLE for single asset removal, COMBINATION for combination of all coins removal
        asset (str)
        shareAmount (float): Mandatory for liquidity removal
    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameters(
        [
            [poolId, "poolId"],
            [type, "type"],
            [asset, "asset"],
            [shareAmount, "shareAmount"],
        ]
    )
    payload = {
        "poolId": poolId,
        "type": type,
        "asset": ",".join(asset),
        "shareAmount": shareAmount,
        **kwargs,
    }

    return self.sign_request("POST", "/sapi/v1/bswap/liquidityRemove", payload)


def bswap_liquidity_operation_record(self, **kwargs):
    """Get Liquidity Operation Record (USER_DATA)
    Get liquidity operation (add/remove) records.

    GET /sapi/v1/bswap/liquidityOps

    https://binance-docs.github.io/apidocs/spot/en/#get-liquidity-operation-record-user_data

    Keyword Args:
        operationId (int, optional)
        poolId (int, optional)
        operation (str, optional): ADD or REMOVE
        startTime (int, optional)
        endTime (int, optional)
        limit (int, optional): default 3, max 100
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    return self.sign_request("GET", "/sapi/v1/bswap/liquidityOps", kwargs)


def bswap_request_quote(
    self, quoteAsset: str, baseAsset: str, quoteQty: float, **kwargs
):
    """Request Quote (USER_DATA)
    Request a quote for swap quote asset (selling asset) for base asset (buying asset), essentially price/exchange rates.
    quoteQty is quantity of quote asset (to sell).

    Please be noted the quote is for reference only, the actual price will change as the liquidity changes,
    it's recommended to swap immediate after request a quote for slippage prevention.

    GET /sapi/v1/bswap/quote

    https://binance-docs.github.io/apidocs/spot/en/#request-quote-user_data

    Args:
        quoteAsset (str)
        baseAsset (str)
        quoteQty (float)
    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameters(
        [[quoteAsset, "quoteAsset"], [baseAsset, "baseAsset"], [quoteQty, "quoteQty"]]
    )
    payload = {
        "quoteAsset": quoteAsset,
        "baseAsset": baseAsset,
        "quoteQty": quoteQty,
        **kwargs,
    }

    return self.sign_request("GET", "/sapi/v1/bswap/quote", payload)


def bswap_swap(self, quoteAsset: str, baseAsset: str, quoteQty: float, **kwargs):
    """Swap (TRADE)
    Swap quoteAsset for baseAsset.

    POST /sapi/v1/bswap/swap

    https://binance-docs.github.io/apidocs/spot/en/#swap-trade

    Args:
        baseAsset (str)
        quoteAsset (str)
        quoteQty (float)
    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameters(
        [[quoteAsset, "quoteAsset"], [baseAsset, "baseAsset"], [quoteQty, "quoteQty"]]
    )
    payload = {
        "quoteAsset": quoteAsset,
        "baseAsset": baseAsset,
        "quoteQty": quoteQty,
        **kwargs,
    }
    return self.sign_request("POST", "/sapi/v1/bswap/swap", payload)


def bswap_swap_history(self, **kwargs):
    """Get Swap History (USER_DATA)
    Get swap history.

    GET /sapi/v1/bswap/swap

    https://binance-docs.github.io/apidocs/spot/en/#get-swap-history-user_data

    Keyword Args:
        swapId (int, optional)
        startTime (int, optional)
        endTime (int, optional)
        status (int, optional)
        baseAsset (str, optional)
        quoteAsset (str, optional)
        limit (int, optional)
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    return self.sign_request("GET", "/sapi/v1/bswap/swap", kwargs)


def bswap_pool_configure(self, **kwargs):
    """Get Pool Configure (USER_DATA)

    GET /sapi/v1/bswap/poolConfigure

    https://binance-docs.github.io/apidocs/spot/en/#get-pool-configure-user_data

    Keyword Args:
        poolId (int, optional)
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    return self.sign_request("GET", "/sapi/v1/bswap/poolConfigure", kwargs)


def bswap_add_liquidity_preview(
    self, poolId: int, type: str, quoteAsset: str, quoteQty: float, **kwargs
):
    """Add Liquidity Preview (USER_DATA)
    Calculate expected share amount for adding liquidity in single or dual token.

    GET /sapi/v1/bswap/addLiquidityPreview

    https://binance-docs.github.io/apidocs/spot/en/#add-liquidity-preview-user_data

    Args:
        poolId (int)
        type (str): "SINGLE" for adding a single token;"COMBINATION" for adding dual tokens
        quoteAsset (str)
        quoteQty (float)
    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameters(
        [
            [poolId, "poolId"],
            [type, "type"],
            [quoteAsset, "quoteAsset"],
            [quoteQty, "quoteQty"],
        ]
    )
    payload = {
        "poolId": poolId,
        "type": type,
        "quoteAsset": quoteAsset,
        "quoteQty": quoteQty,
        **kwargs,
    }
    return self.sign_request("GET", "/sapi/v1/bswap/addLiquidityPreview", payload)


def bswap_remove_liquidity_preview(
    self, poolId: int, type: str, quoteAsset: str, shareAmount: float, **kwargs
):
    """Remove Liquidity Preview (USER_DATA)
    Calculate the expected asset amount of single token redemption or dual token redemption.

    GET /sapi/v1/bswap/removeLiquidityPreview

    https://binance-docs.github.io/apidocs/spot/en/#remove-liquidity-preview-user_data

    Args:
        poolId (int)
        type (str): Type is "SINGLE", remove and obtain a single token;Type is "COMBINATION", remove and obtain dual token
        quoteAsset (str)
        shareAmount (float)
    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameters(
        [
            [poolId, "poolId"],
            [type, "type"],
            [quoteAsset, "quoteAsset"],
            [shareAmount, "shareAmount"],
        ]
    )
    payload = {
        "poolId": poolId,
        "type": type,
        "quoteAsset": quoteAsset,
        "shareAmount": shareAmount,
        **kwargs,
    }
    return self.sign_request("GET", "/sapi/v1/bswap/removeLiquidityPreview", payload)


def bswap_unclaimed_rewards(self, **kwargs):
    """Get Unclaimed Rewards Record (USER_DATA)
    Get unclaimed rewards record.

    GET /sapi/v1/bswap/unclaimedRewards

    https://binance-docs.github.io/apidocs/spot/en/#get-unclaimed-rewards-record-user_data

    Keyword Args:
        type (int, optional): 0: Swap rewards,1:Liquidity rewards, default to 0
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/bswap/unclaimedRewards", kwargs)


def bswap_claim_rewards(self, **kwargs):
    """Claim rewards (TRADE)
    Claim swap rewards or liquidity rewards

    POST /sapi/v1/bswap/claimRewards

    https://binance-docs.github.io/apidocs/spot/en/#claim-rewards-trade

    Keyword Args:
        type (int, optional): 0: Swap rewards,1:Liquidity rewards, default to 0
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("POST", "/sapi/v1/bswap/claimRewards", kwargs)


def bswap_claimed_rewards(self, **kwargs):
    """Get Claimed History (USER_DATA)
    Get history of claimed rewards.

    GET /sapi/v1/bswap/claimedHistory

    https://binance-docs.github.io/apidocs/spot/en/#get-claimed-history-user_data

    Keyword Args:
        poolId (int, optional)
        assetRewards (str, optional)
        type (int, optional): 0: Swap rewards,1:Liquidity rewards, default to 0
        startTime (int, optional)
        endTime (int, optional)
        limit (int, optional): default 3, max 100
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/bswap/claimedHistory", kwargs)



from binance.lib.utils import check_required_parameter
from binance.lib.utils import check_required_parameters
from binance.lib.utils import check_enum_parameter
from binance.lib.enums import TransferType


def system_status(self):
    """System Status (System)
    Fetch system status.

    GET /sapi/v1/system/status

    https://binance-docs.github.io/apidocs/spot/en/#system-status-sapi-system
    """

    return self.query("/sapi/v1/system/status")


def coin_info(self, **kwargs):
    """All Coins' Information (USER_DATA)
    Get information of coins (available for deposit and withdraw) for user.

    GET /sapi/v1/capital/config/getall

    https://binance-docs.github.io/apidocs/spot/en/#all-coins-39-information-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/capital/config/getall", kwargs)


def account_snapshot(self, type: str, **kwargs):
    """Daily Account Snapshot (USER_DATA)

    GET /sapi/v1/accountSnapshot

    https://binance-docs.github.io/apidocs/spot/en/#daily-account-snapshot-user_data

    Parameteres:
    type -- mandatory/string -- "SPOT", "MARGIN", "FUTURES"

    Args:
        type (str): "SPOT", "MARGIN", "FUTURES"
    Keyword Args:
        startTime (int, optional)
        endTime (int, optional)
        limit (int, optional): min 7, max 30, default 7
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameter(type, "type")
    payload = {"type": type, **kwargs}
    return self.sign_request("GET", "/sapi/v1/accountSnapshot", payload)


def disable_fast_withdraw(self, **kwargs):
    """Disable Fast Withdraw Switch (USER_DATA)

    POST /sapi/v1/account/disableFastWithdrawSwitch

    https://binance-docs.github.io/apidocs/spot/en/#disable-fast-withdraw-switch-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request(
        "POST", "/sapi/v1/account/disableFastWithdrawSwitch", kwargs
    )


def enable_fast_withdraw(self, **kwargs):
    """Enable Fast Withdraw Switch (USER_DATA)

    POST /sapi/v1/account/enableFastWithdrawSwitch

    https://binance-docs.github.io/apidocs/spot/en/#enable-fast-withdraw-switch-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request(
        "POST", "/sapi/v1/account/enableFastWithdrawSwitch", kwargs
    )


def withdraw(self, coin: str, amount: float, address: str, **kwargs):
    """Withdraw (USER_DATA)
    Submit a withdraw request.

    POST /sapi/v1/capital/withdraw/apply

    https://binance-docs.github.io/apidocs/spot/en/#withdraw-user_data

    Args:
        coin (str)
        amount (float)
        address (str)
    Keyword Args:
        withdrawOrderId (str, optional): Client id for withdraw
        network (str, optional)
        addressTag (str, optional): Secondary address identifier for coins like XRP,XMR etc.
        transactionFeeFlag (bool, optional): When making internal transfer, True for returning the fee to the destination account; False for returning the fee back to the departure account. Default False.
        name (str, optional): Description of the address. Space in name should be encoded into %20.
        walletType (int, optional): The wallet type for withdraw，0-spot wallet，1-funding wallet. Default is spot wallet
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameters(
        [[coin, "coin"], [amount, "amount"], [address, "address"]]
    )
    payload = {"coin": coin, "amount": amount, "address": address, **kwargs}
    return self.sign_request("POST", "/sapi/v1/capital/withdraw/apply", payload)


def deposit_history(self, **kwargs):
    """Deposit History（supporting network） (USER_DATA)
    Fetch deposit history.

    GET /sapi/v1/capital/deposit/hisrec

    https://binance-docs.github.io/apidocs/spot/en/#deposit-history-supporting-network-user_data

    Keyword Args:
        coin (str, optional)
        status (int, optional): Default 0 (0:pending, 6: credited but cannot withdraw, 1:success)
        startTime (int, optional): Default: 90 days from current timestamp
        endTime (int, optional): Default: present timestamp
        offset (int, optional): Default value: 0
        limit (int, optional): Default:1000, Max:1000
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/capital/deposit/hisrec", kwargs)


def withdraw_history(self, **kwargs):
    """Withdraw History (supporting network) (USER_DATA)
    Fetch withdraw history.

    GET /sapi/v1/capital/withdraw/history

    https://binance-docs.github.io/apidocs/spot/en/#deposit-history-user_data

    Keyword Args:
        coin (str, optional)
        withdrawOrderId (str, optional)
        status (int, optional): Default 0 (0:Email Sent,1:Cancelled, 2:Awaiting Approval,
                3:Rejected, 4:Processing, 5:Failure, 6:Completed)
        offset (int, optional)
        limit (int, optional): Default: 1000, Max: 1000
        startTime (int, optional): Default: 90 days from current timestamp
        endTime (int, optional): Default: present timestamp
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/capital/withdraw/history", kwargs)


def deposit_address(self, coin: str, **kwargs):
    """Deposit Address (supporting network) (USER_DATA)
    Fetch deposit address with network.

    GET /sapi/v1/capital/deposit/address

    https://binance-docs.github.io/apidocs/spot/en/#deposit-address-supporting-network-user_data

    Keyword Args:
        coin (str, optional)
        network (str, optional)
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameter(coin, "coin")
    payload = {"coin": coin, **kwargs}
    return self.sign_request("GET", "/sapi/v1/capital/deposit/address", payload)


def account_status(self, **kwargs):
    """Account Status (USER_DATA)
    Fetch account status detail.

    GET /sapi/v1/account/status

    https://binance-docs.github.io/apidocs/spot/en/#account-status-sapi-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/account/status", kwargs)


def api_trading_status(self, **kwargs):
    """Account API Trading Status (USER_DATA)
    Fetch account api trading status detail.

    GET /sapi/v1/account/apiTradingStatus

    https://binance-docs.github.io/apidocs/spot/en/#account-api-trading-status-sapi-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/account/apiTradingStatus", kwargs)


def dust_log(self, **kwargs):
    """DustLog (USER_DATA)
    Fetch small amounts of assets exchanged BNB records.

    GET /sapi/v1/asset/dribblet

    https://binance-docs.github.io/apidocs/spot/en/#dustlog-sapi-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/asset/dribblet", kwargs)


def user_universal_transfer(self, type: str, asset: str, amount: str, **kwargs):
    """User Universal Transfer (USER_DATA)

    POST /sapi/v1/asset/transfer

    https://binance-docs.github.io/apidocs/spot/en/#user-universal-transfer-user_data

    Args:
        type (str)
        asset (str)
        amount (str)
    Keyword Args:
        fromSymbol (str, optional): Must be sent when type are ISOLATEDMARGIN_MARGIN and ISOLATEDMARGIN_ISOLATEDMARGIN
        toSymbol (str, optional): Must be sent when type are MARGIN_ISOLATEDMARGIN and ISOLATEDMARGIN_ISOLATEDMARGIN
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameters([[type, "type"], [asset, "asset"], [amount, "amount"]])
    check_enum_parameter(type, TransferType)
    payload = {"type": type, "asset": asset, "amount": amount, **kwargs}
    return self.sign_request("POST", "/sapi/v1/asset/transfer", payload)


def user_universal_transfer_history(self, type: str, **kwargs):
    """Query User Universal Transfer History (USER_DATA)

    GET /sapi/v1/asset/transfer

    https://binance-docs.github.io/apidocs/spot/en/#query-user-universal-transfer-history-user_data

     Args:
        type (str)
     Keyword Args:
        startTime (int, optional)
        endTime (int, optional)
        current (int, optional): Default 1
        size (int, optional): Default 10, Max 100
        fromSymbol (str, optional): Must be sent when type are ISOLATEDMARGIN_MARGIN and ISOLATEDMARGIN_ISOLATEDMARGIN
        toSymbol (str, optional): Must be sent when type are MARGIN_ISOLATEDMARGIN and ISOLATEDMARGIN_ISOLATEDMARGIN
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameter(type, "type")
    check_enum_parameter(type, TransferType)
    payload = {"type": type, **kwargs}
    return self.sign_request("GET", "/sapi/v1/asset/transfer", payload)


def transfer_dust(self, asset, **kwargs):
    """Dust Transfer (USER_DATA)
    Convert dust assets to BNB.

    POST /sapi/v1/asset/dust

    https://binance-docs.github.io/apidocs/spot/en/#dust-transfer-user_data

    Args:
        asset (str)
    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameter(asset, "asset")
    payload = {"asset": asset, **kwargs}

    return self.sign_request("POST", "/sapi/v1/asset/dust", payload)


def asset_dividend_record(self, **kwargs):
    """Asset Dividend Record (USER_DATA)
    Query asset dividend record.

    GET /sapi/v1/asset/assetDividend

    https://binance-docs.github.io/apidocs/spot/en/#asset-dividend-record-user_data

    Keyword Args:
        asset (str, optional)
        startTime (int, optional)
        endTime (int, optional)
        limit (int, optional): Default 20, max 500
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/asset/assetDividend", kwargs)


def asset_detail(self, **kwargs):
    """Asset Detail (USER_DATA)
    Fetch details of assets supported on Binance.

    GET /sapi/v1/asset/assetDetail

    https://binance-docs.github.io/apidocs/spot/en/#asset-detail-sapi-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/asset/assetDetail", kwargs)


def trade_fee(self, **kwargs):
    """Trade Fee (USER_DATA)
    Fetch trade fee, values in percentage.

    GET /sapi/v1/asset/tradeFee

    https://binance-docs.github.io/apidocs/spot/en/#trade-fee-sapi-user_data

    Keyword Args:
        symbol (str, optional)
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/asset/tradeFee", kwargs)


def funding_wallet(self, **kwargs):
    """Funding Wallet (USER_DATA)

    POST /sapi/v1/asset/get-funding-asset

    https://binance-docs.github.io/apidocs/spot/en/#funding-wallet-user_data

    Keyword Args:
        asset (str, optional)
        needBtcValuation (str, optional): true or false
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("POST", "/sapi/v1/asset/get-funding-asset", kwargs)


def user_asset(self, **kwargs):
    """User Asset (USER_DATA)

    Get user assets, just for positive data.

    Weight(IP): 5

    POST /sapi/v3/asset/getUserAsset

    https://binance-docs.github.io/apidocs/spot/en/#user-asset-user_data

    Keyword Args:
        asset (str, optional): If asset is blank, then query all positive assets user have.
        needBtcValuation (str, optional)
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    url_path = "/sapi/v3/asset/getUserAsset"
    return self.sign_request("POST", url_path, {**kwargs})


def api_key_permissions(self, **kwargs):
    """Get API Key Permission (USER_DATA)

    GET /sapi/v1/account/apiRestrictions

    https://binance-docs.github.io/apidocs/spot/en/#get-api-key-permission-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/account/apiRestrictions", kwargs)


def bnb_convertible_assets(self, **kwargs):
    """Get Assets That Can Be Converted Into BNB (USER_DATA)

    POST /sapi/v1/asset/dust-btc

    https://binance-docs.github.io/apidocs/spot/en/#get-assets-that-can-be-converted-into-bnb-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("POST", "/sapi/v1/asset/dust-btc", kwargs)


def convertible_coins(self, **kwargs):
    """Query auto-converting stable coins (USER_DATA)

    GET /sapi/v1/capital/contract/convertible-coins

    https://binance-docs.github.io/apidocs/spot/en/#query-auto-converting-stable-coins-user_data
    """

    return self.sign_request(
        "GET", "/sapi/v1/capital/contract/convertible-coins", kwargs
    )


def toggle_auto_convertion(self, coin: str, enable: bool, **kwargs):
    """Toggle auto-converting stable coins (USER_DATA)

    POST /sapi/v1/capital/contract/convertible-coins

    https://binance-docs.github.io/apidocs/spot/en/#switch-on-off-busd-and-stable-coins-conversion-user_data
    """

    check_required_parameters([[coin, "symbol"], [enable, "enable"]])

    payload = {"coin": coin, "enable": enable, **kwargs}
    return self.sign_request(
        "POST", "/sapi/v1/capital/contract/convertible-coins", payload
    )


def cloud_mining_trans_history(self, startTime: int, endTime: int, **kwargs):
    """Get Cloud-Mining payment and refund history (USER_DATA)

    GET /sapi/v1/asset/ledger-transfer/cloud-mining/queryByPage

    https://binance-docs.github.io/apidocs/spot/en/#get-cloud-mining-payment-and-refund-history-user_data

    Args:
        startTime (int)
        endTime (int)
    Keyword Args:
        tranId (int, optional)
        clientTranId (str, optional)
        asset (str, optional)
        current (int, optional): Default Value: 1
        size (int, optional): Default Value: 100; Max Value: 100
        recvWindow (int, optional)
    """

    check_required_parameters([[startTime, "startTime"], [endTime, "endTime"]])

    url_path = "/sapi/v1/asset/ledger-transfer/cloud-mining/queryByPage"
    payload = {"startTime": startTime, "endTime": endTime, **kwargs}
    return self.sign_request("GET", url_path, payload)


def convert_transfer(
    self, clientTranId: str, asset: str, amount: float, targetAsset: str, **kwargs
):
    """BUSD Convert (USER_DATA)

    POST /sapi/v1/asset/convert-transfer

    https://binance-docs.github.io/apidocs/spot/en/#busd-convert-trade

    Args:
        clientTranId (str)
        asset (str)
        amount (float)
        targetAsset (str)
    Keyword Args:
        accountType (str, optional)
        recvWindow (int, optional)
    """

    check_required_parameters(
        [
            [clientTranId, "clientTranId"],
            [asset, "asset"],
            [amount, "amount"],
            [targetAsset, "targetAsset"],
        ]
    )

    url_path = "/sapi/v1/asset/convert-transfer"
    payload = {
        "clientTranId": clientTranId,
        "asset": asset,
        "amount": amount,
        "targetAsset": targetAsset,
        **kwargs,
    }
    return self.sign_request("POST", url_path, payload)


def convert_history(self, startTime: int, endTime: int, **kwargs):
    """BUSD Convert History (USER_DATA)

    GET /sapi/v1/asset/convert-transfer/queryByPage

    https://binance-docs.github.io/apidocs/spot/en/#busd-convert-history-user_data

    Args:
        startTime (int)
        endTime (int)
    Keyword Args:
        tranId (int, optional)
        clientTranId (str, optional)
        asset (str, optional)
        accountType (str, optional)
        current (int, optional): Default Value: 1
        size (int, optional): Default Value: 100; Max Value: 100
        recvWindow (int, optional)
    """

    check_required_parameters([[startTime, "startTime"], [endTime, "endTime"]])

    url_path = "/sapi/v1/asset/convert-transfer/queryByPage"
    payload = {"startTime": startTime, "endTime": endTime, **kwargs}
    return self.sign_request("GET", url_path, payload)
